		cblock	0x55
		ACCaLO
		ACCaHI
		ACCbLO
		ACCbHI
		ACCcLO
		ACCcHI
		ACCdLO
		ACCdHI
		temp
		sign
		Flags
		endc
;
TRUE    equ     1
FALSE   equ     0
MSB     equ     7
;*******************************************************************
SIGNED  equ     FALSE	        ; Set This To eTRUEi if the routines
;                               ; for Multiplication & Division needs
;                               ; to be assembled as Signed Integer
;                               ; Routines. If eFALSEi the above two
;                               ; routines ( D_mpy & D_div ) use
;                               ; unsigned arithmetic.
;*******************************************************************
;                       Double Precision Division
;
;               ( Optimized for Code Size : Looped Code )
;
;*******************************************************************;
; Division : ACCb(16 bits) / ACCa(16 bits) -> ACCb(16 bits) with
;                                          Remainder in ACCc (16 bits)
;    (a) Load the Denominator in location ACCaHI & ACCaLO ( 16 bits )
;    (b) Load the Numerator in location ACCbHI & ACCbLO ( 16 bits )
;    (c) CALL D_div
;    (d) The 16 bit result is in location ACCbHI & ACCbLO
;    (e) The 16 bit Remainder is in locations ACCcHI & ACCcLO
;
;   Performance :
;               Program Memory  :       037
;               Clock Cycles    :       310
;
;        NOTE :
;               The performance specs are for Unsigned arithmetic 
;               ( i.e,with iSIGNED equ  FALSE i).
;
;
;       Program:          DBL_DIVS.ASM 
;       Revision Date:   
;                         1-13-97      Compatibility with MPASMWIN 1.40
;
;*******************************************************************;
;(ACCb/ ACCa -> ACCb   ACCc)
;*******************************************************************
;       Double Precision Divide ( 16/16 -> 16 )
;
; (ACCb/ACCa -> ACCb with remainder in ACCc) : 16 bit output
; with Quotiont in ACCb (ACCbHI,ACCbLO) and Remainder in ACCc 
; (ACCcHI,ACCcLO).
; NOTE: Before calling this routine, the user should make sure that
;       the Numerator(ACCb) is greater than Denominator(ACCa). If
;       the case is not true, the user should scale either Numerator
;       or Denominator or both such that Numerator is greater than
;       the Denominator.
;
;
D_divS
;
     IF   SIGNED
     CALL    S_SIGN
     ENDIF
;
        call    setup
        clrf    ACCcHI
        clrf    ACCcLO
dloop   bcf     STATUS,C
        rlf     ACCdLO, F
        rlf     ACCdHI, F
        rlf     ACCcLO, F
        rlf     ACCcHI, F
        movf    ACCaHI,W
        subwf   ACCcHI,W        ; check if a>c
        btfss   STATUS,Z
        goto    nochk
        movf    ACCaLO,W
        subwf   ACCcLO,W        ; if msb equal then check lsb
nochk   btfss   STATUS,C        ; carry set if c>a
        goto    nogo
        movf    ACCaLO,W        ; c-a into c
        subwf   ACCcLO, F
        btfss   STATUS,C
        decf    ACCcHI, F
        movf    ACCaHI,W
        subwf   ACCcHI, F
        bsf     STATUS,C        ; shift a 1 into b (result)
nogo    rlf     ACCbLO, F
        rlf     ACCbHI, F
        decfsz  temp, F         ; loop untill all bits checked
        goto    dloop
;
     IF    SIGNED
        btfss   sign,MSB        ; check sign if negative
        retlw   0
        goto    neg_B           ; negate ACCa ( -ACCa -> ACCa )
    ELSE
        retlw   0
    ENDIF
;
;*******************************************************************
;               Double Precision Multiply ( 16x16 -> 32 )
;         ( ACCb*ACCa -> ACCb,ACCc ) : 32 bit output with high word
;  in ACCb ( ACCbHI,ACCbLO ) and low word in ACCc ( ACCcHI,ACCcLO ).
;
D_mpyS                           ;results in ACCb(16 msbis) and ACCc(16 lsbis)
;
     IF   SIGNED
     	call    S_SIGN
     ENDIF
;
        call    setup
mloop   rrf     ACCdHI, F       ;rotate d right
        rrf     ACCdLO, F
        btfsc   STATUS, C        ;need to add?
        call    D_add
        rrf     ACCbHI, F
        rrf     ACCbLO, F
        rrf     ACCcHI, F
        rrf     ACCcLO, F
        decfsz  temp, F         ;loop until all bits checked
        goto    mloop
;
    IF    SIGNED
        btfss   sign,MSB
        retlw   0
        comf    ACCcLO, F       ; negate ACCa ( -ACCa -> ACCa )
        incf    ACCcLO, F
        btfsc   STATUS,Z
        decf    ACCcHI, F
        comf    ACCcHI, F
        btfsc   STATUS,Z
neg_B   comf    ACCbLO, F       ; negate ACCb
        incf    ACCbLO, F
        btfsc   STATUS,Z
        decf    ACCbHI, F
        comf    ACCbHI, F
        retlw   0
    ELSE
        retlw   0
    ENDIF
;
;*******************************************************************
;
setup   movlw   .16             ; for 16 shifts
        movwf   temp
        movf    ACCbHI,W        ; move ACCb to ACCd
        movwf   ACCdHI
        movf    ACCbLO,W
        movwf   ACCdLO
        clrf    ACCbHI
        clrf    ACCbLO
		retlw   0
;
;*******************************************************************
;
neg_A   comf    ACCaLO, F       ; negate ACCa ( -ACCa -> ACCa )
        incf    ACCaLO, F
        btfsc   STATUS,Z
        decf    ACCaHI, F
        comf    ACCaHI, F
        retlw   0

;*******************************************************************
;  Assemble this section only if Signed Arithmetic Needed
;
     IF    SIGNED
;
S_SIGN  movf    ACCaHI,W
        xorwf   ACCbHI,W
        movwf   sign
        btfss   ACCbHI,MSB        ; if MSB set go & negate ACCb
        goto    chek_A
;
        comf    ACCbLO            ; negate ACCb
        incf    ACCbLO
        btfsc   STATUS,Z
        decf    ACCbHI
        comf    ACCbHI
;
chek_A  btfss   ACCaHI,MSB        ; if MSB set go & negate ACCa
        retlw   0
        goto    neg_A
;
     ENDIF

;*******************************************************************
;                 Double Precision Addition & Subtraction
;
;*******************************************************************;
;   Addition :  ACCb(16 bits) + ACCa(16 bits) -> ACCb(16 bits)
;      (a) Load the 1st operand in location ACCaLO & ACCaHI ( 16 bits )
;      (b) Load the 2nd operand in location ACCbLO & ACCbHI ( 16 bits )
;      (c) CALL D_add
;      (d) The result is in location ACCbLO & ACCbHI ( 16 bits )
;
;   Performance :
;               Program Memory  :  07
;               Clock Cycles    :  08
;*******************************************************************;
;   Subtraction : ACCb(16 bits) - ACCa(16 bits) -> ACCb(16 bits)
;      (a) Load the 1st operand in location ACCaLO & ACCaHI ( 16 bits )
;      (b) Load the 2nd operand in location ACCbLO & ACCbHI ( 16 bits )
;      (c) CALL D_sub
;      (d) The result is in location ACCbLO & ACCbHI ( 16 bits )
;
;   Performance :
;               Program Memory  :       14
;               Clock Cycles    :       17
;
;
;       Program:          DBL_ADD.ASM 
;       Revision Date:   
;                         1-13-97      Compatibility with MPASMWIN 1.40
;
;*******************************************************************;
;*******************************************************************
;         Double Precision Subtraction ( ACCb - ACCa -> ACCb )
;
D_sub   call    neg_A           ; At first negate ACCa; Then add
;
;*******************************************************************
;       Double Precision  Addition ( ACCb + ACCa -> ACCb )
;
D_add   movf    ACCaLO,W
        addwf   ACCbLO, F       ; add lsb
        btfsc   STATUS,C        ; add in carry
        incf    ACCbHI, F
        movf    ACCaHI,W
        addwf   ACCbHI, F       ; add msb
        retlw   0

