		cblock	0x55
		ACCaLO
		ACCaHI
		ACCbLO
		ACCbHI
		ACCcLO
		ACCcHI
		ACCdLO
		ACCdHI
		temp
		sign
		Flags
		endc
;
MSB     equ     7
;*******************************************************************
;                       Double Precision Division
;
;               ( Optimized for Code Size : Looped Code )
;
;*******************************************************************;
; Division : ACCb(16 bits) / ACCa(16 bits) -> ACCb(16 bits) with
;                                          Remainder in ACCc (16 bits)
;    (a) Load the Denominator in location ACCaHI & ACCaLO ( 16 bits )
;    (b) Load the Numerator in location ACCbHI & ACCbLO ( 16 bits )
;    (c) CALL D_div
;    (d) The 16 bit result is in location ACCbHI & ACCbLO
;    (e) The 16 bit Remainder is in locations ACCcHI & ACCcLO
;
;   Performance :
;               Program Memory  :       037
;               Clock Cycles    :       310
;
;        NOTE :
;               The performance specs are for Unsigned arithmetic 
;               ( i.e,with iSIGNED equ  FALSE i).
;
;
;       Program:          DBL_DIVS.ASM 
;       Revision Date:   
;                         1-13-97      Compatibility with MPASMWIN 1.40
;
;*******************************************************************;
;(ACCb/ ACCa -> ACCb   ACCc)
;*******************************************************************
;       Double Precision Divide ( 16/16 -> 16 )
;
; (ACCb/ACCa -> ACCb with remainder in ACCc) : 16 bit output
; with Quotiont in ACCb (ACCbHI,ACCbLO) and Remainder in ACCc 
; (ACCcHI,ACCcLO).
; NOTE: Before calling this routine, the user should make sure that
;       the Numerator(ACCb) is greater than Denominator(ACCa). If
;       the case is not true, the user should scale either Numerator
;       or Denominator or both such that Numerator is greater than
;       the Denominator.
;
;
D_divS
        call    setup
        clrf    ACCcHI
        clrf    ACCcLO
dloop   bcf     STATUS,C
        rlf     ACCdLO, F
        rlf     ACCdHI, F
        rlf     ACCcLO, F
        rlf     ACCcHI, F
        movf    ACCaHI,W
        subwf   ACCcHI,W        ; check if a>c
        btfss   STATUS,Z
        goto    nochk
        movf    ACCaLO,W
        subwf   ACCcLO,W        ; if msb equal then check lsb
nochk   btfss   STATUS,C        ; carry set if c>a
        goto    nogo
        movf    ACCaLO,W        ; c-a into c
        subwf   ACCcLO, F
        btfss   STATUS,C
        decf    ACCcHI, F
        movf    ACCaHI,W
        subwf   ACCcHI, F
        bsf     STATUS,C        ; shift a 1 into b (result)
nogo    rlf     ACCbLO, F
        rlf     ACCbHI, F
        decfsz  temp, F         ; loop untill all bits checked
        goto    dloop
        retlw   0
;
;*******************************************************************
;               Double Precision Multiply ( 16x16 -> 32 )
;         ( ACCb*ACCa -> ACCb,ACCc ) : 32 bit output with high word
;  in ACCb ( ACCbHI,ACCbLO ) and low word in ACCc ( ACCcHI,ACCcLO ).
;
D_mpyS                           ;results in ACCb(16 msbis) and ACCc(16 lsbis)
        call    setup
mloop   rrf     ACCdHI, F       ;rotate d right
        rrf     ACCdLO, F
        btfsc   STATUS, C        ;need to add?
        call    D_add
        rrf     ACCbHI, F
        rrf     ACCbLO, F
        rrf     ACCcHI, F
        rrf     ACCcLO, F
        decfsz  temp, F         ;loop until all bits checked
        goto    mloop
        retlw   0
;
;*******************************************************************
;
setup   movlw   .16             ; for 16 shifts
        movwf   temp
        movf    ACCbHI,W        ; move ACCb to ACCd
        movwf   ACCdHI
        movf    ACCbLO,W
        movwf   ACCdLO
        clrf    ACCbHI
        clrf    ACCbLO
		retlw   0
;
;*******************************************************************
;
neg_A   comf    ACCaLO, F       ; negate ACCa ( -ACCa -> ACCa )
        incf    ACCaLO, F
        btfsc   STATUS,Z
        decf    ACCaHI, F
        comf    ACCaHI, F
        retlw   0
;
;*******************************************************************
;                 Double Precision Addition & Subtraction
;
;*******************************************************************;
;   Addition :  ACCb(16 bits) + ACCa(16 bits) -> ACCb(16 bits)
;      (a) Load the 1st operand in location ACCaLO & ACCaHI ( 16 bits )
;      (b) Load the 2nd operand in location ACCbLO & ACCbHI ( 16 bits )
;      (c) CALL D_add
;      (d) The result is in location ACCbLO & ACCbHI ( 16 bits )
;
;   Performance :
;               Program Memory  :  07
;               Clock Cycles    :  08
;*******************************************************************;
;   Subtraction : ACCb(16 bits) - ACCa(16 bits) -> ACCb(16 bits)
;      (a) Load the 1st operand in location ACCaLO & ACCaHI ( 16 bits )
;      (b) Load the 2nd operand in location ACCbLO & ACCbHI ( 16 bits )
;      (c) CALL D_sub
;      (d) The result is in location ACCbLO & ACCbHI ( 16 bits )
;
;   Performance :
;               Program Memory  :       14
;               Clock Cycles    :       17
;
;
;       Program:          DBL_ADD.ASM 
;       Revision Date:   
;                         1-13-97      Compatibility with MPASMWIN 1.40
;
;*******************************************************************;
;*******************************************************************
;         Double Precision Subtraction ( ACCb - ACCa -> ACCb )
;
D_sub   call    neg_A           ; At first negate ACCa; Then add
;
;*******************************************************************
;       Double Precision  Addition ( ACCb + ACCa -> ACCb )
;
D_add   movf    ACCaLO,W
        addwf   ACCbLO, F       ; add lsb
        btfsc   STATUS,C        ; add in carry
        incf    ACCbHI, F
        movf    ACCaHI,W
        addwf   ACCbHI, F       ; add msb
        retlw   0

;
;*******************************************************************
;
;            Square Root By Newton Raphson Method
;
;  This routine computes the square root of a 16 bit number(with
;  low byte in NumLo & high byte in NumHi ). After loading NumLo &
;  NumHi with the desired number whose square root is to be computed,
;  branch to location Sqrt ( by iGOTO  Sqrti ). i CALL  Sqrti cannot
;  be issued because the Sqrt function makes calls to Math routines
;  and the stack is completely used up.
;  The result = sqrt(NumHi,NumLo) is returned in location SqrtLo.
;  The total number of iterations is set to ten. If more iterations
;  are desired, change iLupCnt equ .10i to the desired value. Also,
;  the initial guess value of the square root is given set as
;  input/2 ( in subroutine iiniti ). The user may modify this scheme
;  if a better initial approximation value is known. A good initial
;  guess will help the algorithm converge at a faster rate and thus
;  less number of iterations required.
;  Two utility math routines are used by this program : D_divS
;  and D_add. These two routines are listed as seperate routines
;  under double precision Division and double precision addtion
;  respectively.
;
;  Note : If square root of an 8 bit number is desired, it is probably
;         better to have a table look scheme rather than using numerical
;         methods.
;
;
;
;   Performance :
;               Program Memory  :       27  (excluding Math Routines
;                                            D_divS & D_add )
;               Clock Cycles    :       3600 ( approximately )
;
;
;       Program:          SQRT.ASM 
;       Revision Date:   
;                         1-13-97      Compatibility with MPASMWIN 1.40
;
;       To assemble this program, two routines, namely iD_addi &
;       iD_divSi must be included into this program. These two routines
;       are listed as separate programs in files iDBL_ADD.ASMi &
;       iDBL_DIVS.ASMi respectively.
;
;*******************************************************************
;
LupCnt  equ     .10             ; Number of iterations


SqrtLo  equ     ACCaLO
SqrtHi  equ     ACCaHI
;
NumLo   equ     0x53
NumHi   equ     0x54
count   equ     Count
;
;
init
        movlw   LupCnt
        movwf   count
        movf    NumHi,W
        movwf   SqrtHi
        movf    NumLo,W         ; set initial guess root = NUM/2
        movwf   SqrtLo
        bcf     STATUS,C
        rrf     SqrtHi, F
        rrf     SqrtLo, F
        retlw   0
;
div2    bcf     STATUS,C
        rrf     ACCbHI,W
        movwf   SqrtHi
        rrf     ACCbLO,W
        movwf   SqrtLo
        retlw   0

Sqrt    call    init
sloop   movf    NumLo,W
        movwf   ACCbLO
        movf    NumHi,W
        movwf   ACCbHI
;
        call    D_divS          ; double precision division
        call    D_add           ; double precision addition
;                               ; the above 2 routines are listed
;                               ; as seperate routines
        call    div2
        decfsz  count, F
        goto    sloop
        goto    over            ; all iterations done
;                               ; branch back to desired location
;
;*************************************************************
;               Test Program
;*************************************************************
;
;main    movlw   0xF3
;        movwf   NumHi
;        movlw   0xF6       ; Set input test number = 62454
;        movwf   NumLo     ;  = F3F6h
;
;        goto    Sqrt      ; cannot use CALL : Math routines
;                         ; use up all the stack.
;over    nop               ; all iterations done
;
;self    goto    self      ; result = 00F9h = 249
;                         ; exact sqrt(62454) = 249.9